/*
  Copyright 2009 CNRS-LAAS

  Author: Oussama Kanoun
*/

#ifndef HPP_LOS_CENTIPEDE_H
#define HPP_LOS_CENTIPEDE_H

#include "MatrixAbstractLayer/MatrixAbstractLayer.h"
#include "hpplocalstepper/hppLosLink.hh"
#include <list>

/**
A serial assemly of ChppLosLink objects. A ChppLosCentipede object is constructed by giving an initial left footprint and an initial right footprint.
\ingroup parts
*/
class ChppLosCentipede
{
    public:
        /**
        \name Instance construction
        @{
         */
        ChppLosCentipede ( const ChppLosFootprint& initialleftprint, const ChppLosFootprint& initialrightprint );
        ~ChppLosCentipede();
        const ChppLosFootprint& initialLeftFootprint();
        const ChppLosFootprint& initialRightFootprint();
        void initialLeftFootprint ( const ChppLosFootprint& inF );
        void initialRightFootprint ( const ChppLosFootprint& inF );
        /**
        @}
         */

        /**
        \name Assembly
        @{
        */
        bool push_back ( ChppLosLink& inLink );
        bool push_front ( ChppLosLink& inLink );
        void pop_back ( );
        void pop_front ( );
        unsigned int numberLinks();
        void clearLinks();
        ChppLosLink* linkAtRank ( unsigned int inRank );
        std::list<ChppLosSupportPolygonLink*>& supportPolygonLinks(){return attSupportPolygonLinks;};
        std::list<ChppLosRobotLink*>& robotLinks(){return attRobotLinks;};
        /**
        @}
        */

        /**
        \name Configuration
        @{
         */
        unsigned int numberDof();
        bool getConfiguration(vectorN& outConfig);
        bool applyConfiguration ( const vectorN& inVector, bool computeRelativeFeetJacobians = true);
        /**
        @}
         */

        /**
        \name Velocity bounds
        @{
         */
        bool getUpperVelocityBounds(vectorN& outConfig);
        bool getLowerVelocityBounds(vectorN& outConfig);
        /**
        @}
         */

        /**
        \name Jacobians
        @{
         */
        bool getJacobian ( unsigned int startLinkRank,unsigned int effectorLinkRank, const vector3d& inAbsPoint, matrixNxP& outJacobian );
        /**
        @}
         */

    private:
        ChppLosFootprint attInitialLeft;
        ChppLosFootprint attInitialRight;

        std::list<ChppLosLink*> attLinks;
        std::list<ChppLosSupportPolygonLink*> attSupportPolygonLinks;
        std::list<ChppLosRobotLink*> attRobotLinks;
        unsigned int attNumberDof;
        vectorN attConfiguration,attUb,attLb, linkConfig;
       
};


#endif
