/*
  Copyright 2009 CNRS-LAAS

  Author: Oussama Kanoun
*/

#ifndef HPP_LOS_SOLVER_H
#define HPP_LOS_SOLVER_H

#include <iostream>
#include "MatrixAbstractLayer/MatrixAbstractLayer.h"
class ChppLosSolverBasic;
class ChppLosTask;
class ChppLosCentipede;


/**
\brief This object solves a stack of prioritized equality and inequality tasks applied to a ChppLosCentipede object.
\ingroup solvers
  */
class ChppLosSolver
{
    public:

        /**
        \brief Constructor.
         */
        ChppLosSolver ( ChppLosCentipede& inCentipede );
        /**
        \brief Set the constraints of the problem.
        */
        void constraints ( const std::vector<ChppLosTask*>& inConstraints );

        /**
        \brief Set the prioritized tasks to be solved. First task in the vector has highest priority.
        */
        void tasks ( const std::vector<ChppLosTask*>& inTasks );

        /**
        \brief Set the joints used in solving. Size of vector must be equal to Centipede's number of dofs
         */
        bool setMask ( const vectorN& inActiveDofs );
        
        /**\brief Solve.
        */
        bool solve(bool savemotionoption= false, const std::string& inFilename = std::string("solution"));
       
        /**\brief Get last configuration reached after call of solve()
         */
        const vectorN& reachedConfiguration() {return attSolution;};

        /**
            \brief Destructor
         */
        ~ChppLosSolver();

    private:
        bool solveIteration();

        void recordGaps();
        bool smallerGaps();
        bool isProblemSolved();
        bool constraintsSatisfied();

        ChppLosCentipede* attCentipede;
        ChppLosSolverBasic* attLosSolver;
        std::vector<ChppLosTask*> attTasks,attConstraints;
        vectorN attSolvedTasks;

        double attSolveThreshold, attProgressThreshold;
        vectorN attActivatedJoints, veloSolution, attSolution,curConfig;
        
        std::vector<double> attTaskGaps;
        
        unsigned int attMaxIter;
};

#endif
