/*
  Copyright 2009 CNRS-LAAS

  Author: Oussama Kanoun
*/

#ifndef HRP2_BASIC_ANTICOLLISION_H
#define HRP2_BASIC_ANTICOLLISION_H

#include "hpplocalstepper/tasks/hppLosTask.hh"
#include "hpplocalstepper/tasks/hppLosCollisionAvoider.hh"
/**
A task for HRP2 to avoid having the arms collide with the trunc or the thighs and both hand from approching the head
\ingroup tasks
 */
class HRP2BasicCollisionAvoider: public ChppLosTask
{
    public:
       /**
        \brief Constructor
        */
        HRP2BasicCollisionAvoider(ChppLosCentipede& inCentipede, ChppLosRobotLink& inLink);
        ~HRP2BasicCollisionAvoider();
        /**
        \brief Set the area of translation brake inside the authorized configurations region boudaries
         */
        void setMargins ( double inDi, double inDs, double inXi );
        /**
        \brief Get the task for avoiding collision of bounding regions
        */
        ChppLosCollisionAvoider& pairCollisionAvoider( ){return *attCA;}
        /**
        \brief Inherited.
         */
        bool isEqualityTask() const {return false;};
        /**
        \brief Inherited.
         */
        void computeConstraint();
        /**
        \brief Inherited.
         */
        const vectorN& currentValue();
        
    private:
        
        std::vector< LosRobotSphere> attRobotSpheres;
        std::vector< LosRobotCylinder> attRobotCylinders;
                
        void calcChestConstraint(const vector3d& inlocchest,unsigned int rank);
        void calcChestValue(const vector3d& inlocchest,unsigned int rank);
        
        ChppLosRobotLink* attLink;
        CjrlJoint* attWaist,*attChest;
        CjrlHumanoidDynamicRobot* robot;
        double attXi,attDi,attDs;
        double attChestzll;
        vector3d ntonext,nout,otop,waist,chest;
        matrix4d m;
        matrixNxP tempJac3d,noutM,jac1d;
        vector3d locChest0,locChest1,chest0,chest1;
        
        ChppLosCollisionAvoider* attCA;
};

#endif
