/*
  Copyright 2009 CNRS-LAAS

  Author: Oussama Kanoun
*/

#ifndef HPP_LOS_DISC_AVOIDER_H
#define HPP_LOS_DISC_AVOIDER_H

#include "hpplocalstepper/tasks/hppLosTask.hh"

/**
A task for a robot to avoid stepping inside a disc
\ingroup tasks
 */
class ChppLosDiscAvoider: public ChppLosTask
{
    public:
       /**
        \brief Constructor
        */
        ChppLosDiscAvoider(ChppLosCentipede& inCentipede);
        /**
        \brief Set the radius of disc region
         */
        void setMargins ( double inDi, double inDs, double inXi );
        /**
        \brief Set the position of the center of the disc region
         */
        void setPosition ( double inX, double inY);
        /**
        \brief Inherited.
         */
        bool isEqualityTask() const {return false;};
        /**
        \brief Inherited.
         */
        void computeConstraint();
        /**
        \brief Inherited.
         */
        const vectorN& currentValue();
        
    private:
        bool resize();
        
        void calcRegionConstraint(ChppLosLink* inLink, unsigned int rank);
        
        void calcRegionValue ( ChppLosLink* inLink, unsigned int rank);
        
        void projectObstacleOnFoot( ChppLosLink* inLink, vector3d& constrainedPoint);

        double attXi,attDi,attDs,attPositionX,attPositionY;
        vector3d otop,proj;
        matrix4d m;
        matrixNxP tempJac3d,noutM,jac1d;
};

#endif
