/*
  Copyright 2009 CNRS-LAAS

  Author: Oussama Kanoun
*/

#ifndef HPP_LOS_POINTING_TASK_H
#define HPP_LOS_POINTING_TASK_H

#include "hpplocalstepper/tasks/hppLosTask.hh"

/**
A task to align a vector in a robot's joint frame on a point in world frame
\ingroup tasks
 */
class ChppLosPointingTask: public ChppLosTaskOnJoint
{
    public:

        /**
        \brief Constructor
         */
        ChppLosPointingTask ( ChppLosCentipede& inCentipede, ChppLosRobotLink& inRobotLink, CjrlJoint& inJoint, const vector3d& inLocalOrigin, const vector3d& inLocalVector, const vector3d& inWorlTarget );
        /**
        \brief Set reference joint
         */
        void joint ( CjrlJoint& inJoint );
        /**
        \brief Set the origin of the pointing vector in joint's local frame (illelgal operation for gaze constraint).
         */
        virtual void  localOrigin ( const vector3d& inPoint ) ;
        /**
        \brief Get the origin of the pointing vector.
         */
        virtual const vector3d& localOrigin() ;
        /**
        \brief Set the pointing vector in joint's local frame (illelgal operation for gaze constraint).
         */
        virtual void  localVector ( const vector3d& inPoint ) ;
        /**
        \brief Set the pointing vector in joint's local frame
         */
        virtual const vector3d& localVector();
        /**
        \brief Set the target point associated to the constraint (in world's frame).
         */
        void  worldTarget ( const vector3d& inPoint );
        /**
        \brief Get the target point associated to the constraint (in world's frame).
         */
        const vector3d& worldTarget();
        /**
        \brief Set the gain for differential equation when solving this tasks
         */
        void maxVelocityNorm ( double inMV );
        /**
        \brief Return true
         */
        bool isEqualityTask() const {return true;};
        /**
        \brief Inherited
         */
        void computeConstraint();
        /**
        \brief Inherited
         */
        const vectorN& currentValue();
        

    protected:
        bool resize();
        void Rodrigues(const vector3d& inW, double inAngle, matrixNxP& outRot);
        void equivAsymMat3(const vector3d& inVec, matrixNxP& outMat);
        void crossProd3(const vector3d& v1, const vector3d& v2, vector3d& retV);
        vector3d attCurrentPoint;
        ChppLosRobotLink* attRobotLink;
        vector3d attLocalOrigin,attLocalVector,attWorldTarget;
        double attMaxVelocity;
        matrixNxP attTempJacobian;
        matrix4d m;
};

#endif
