/*
  Copyright 2009 CNRS-LAAS

  Author: Oussama Kanoun
*/

#ifndef HPP_LOS_POSITION_TASK_H
#define HPP_LOS_POSITION_TASK_H

#include "hpplocalstepper/tasks/hppLosTask.hh"

/**
A 3D position task for a point located in the local frame of a given joint inside a CjrlDynamicHumanoidRobot inside a ChppLosCentipede object.
\ingroup tasks
 */
class ChppLosPositionTask: public ChppLosTaskOnJoint
{
    public:
       
        /**
        \brief Constructor
         */
        ChppLosPositionTask(ChppLosCentipede& inCentipede, ChppLosRobotLink& inRobotLink, CjrlJoint& inJoint, const vector3d& inLocalPoint, const vector3d& inWorlTarget);
        /**
        \brief Get reference point in reference joint
         */
        const vector3d& localPoint();
        /**
        \brief Get target in world coordinates
         */
        const vector3d& targetPoint();
        /**
        \brief Set reference point in reference joint local coordinates
         */
        void localPoint(const vector3d& inVec);
        /**
        \brief Set target point in world coordinates
         */
        void targetPoint(const vector3d& inVec);
        /**
        \brief Set the maximum linear velocity of the reference point
         */
        void maxVelocityNorm(double inMV);
        /**
        \brief Return true
         */
        bool isEqualityTask() const {return true;};
        /**
        \brief Inherited
         */
        void computeConstraint();
        /**
        \brief Inherited
         */
        const vectorN& currentValue();

    private:
        
        vector3d attCurrentPoint;
        ChppLosRobotLink* attRobotLink;
        vector3d attLocalPoint,attWorldTarget;
        double attMaxVelocity;
        matrix4d m;
};

#endif
