/*
  Copyright 2009 CNRS-LAAS

  Author: Oussama Kanoun
*/

#ifndef HPP_LOS_TRANFORMATION_TASK_H
#define HPP_LOS_TRANFORMATION_TASK_H

#include "hpplocalstepper/tasks/hppLosTask.hh"

/**
A task to achieve a given transformation in world coordinates.
\ingroup tasks
 */
class ChppLosTransformationTask: public ChppLosTaskOnJoint
{
    public:

        /**
        \brief Constructor
         */
        ChppLosTransformationTask ( ChppLosCentipede& inCentipede, ChppLosRobotLink& inRobotLink, CjrlJoint& inJoint, const vector3d& inLocalPoint, const matrix4d& inTargetTransform );
        /**
        \brief Get reference point in reference joint
         */
        const vector3d& localPoint();
        /**
        \brief Get target in world coordinates
         */
        const matrix4d& targetTransformation();
        /**
        \brief Set reference point in reference joint local coordinates
         */
        void localPoint ( const vector3d& inVec );
        /**
        \brief Set target transformation in world coordinates
         */
        void targetTransformation ( const matrix4d& inVec );
        /**
        \brief Set the maximum translation velocity of the reference point
         */
        void maxTranslationVelocity ( double inMV );
        /**
        \brief Set the maximum rotation velocity of the reference point
         */
        void maxRotationVelocity ( double inMV );
        /**
        \brief Return true
         */
        bool isEqualityTask() const {return true;};
        /**
        \brief Inherited
         */
        void computeConstraint();
        /**
        \brief Inherited
         */
        const vectorN& currentValue();

    private:
        
        vector3d attCurrentPoint;
        ChppLosRobotLink* attRobotLink;
        vector3d attLocalPoint;
        matrix4d attTargetTransform;
        double attMaxTVelocity,attMaxRVelocity;
        matrixNxP attTempJacobian;
        matrix4d m;
        vectorN tempOmega,CurPos,TarPos,tempGap;
        matrixNxP CurRot,TarRot,tempGapRot;
};

#endif
