/*
  Copyright 2009 CNRS-LAAS

  Author: Oussama Kanoun
*/

#include "hpplocalstepper/hppLosCentipedeFactory.hh"


using namespace boost::numeric::ublas;

ChppLosCentipede* ChppLosCentipedeFactory::createSpringman ( const ChppLosFootprint& initialleftprint, const ChppLosFootprint& initialrightprint, ChppGikStandingRobot& inStandingRobot, unsigned int numberExtraVirtualStepLinks, bool rootIsRight, const ChppLosFootprintArea& leftStepArea,  const ChppLosFootprintArea& righttStepArea, const ChppLosFootprintShape& inLeftFootShape, const ChppLosFootprintShape& inRightFootShape, std::list<ChppLosLink*>& outCentipedeParts,vectorN& outMask )
{

    ChppLosCentipede* outCentipede = new ChppLosCentipede ( initialleftprint,initialrightprint );

    unsigned int ndof = 0;
    vectorN config;
    vectorN sprConfig ( 3 );
    vectorN splConfig ( 3 );

    double drlth = initialleftprint.th-initialrightprint.th;
    double drlx = initialleftprint.x-initialrightprint.x;
    double drly = initialleftprint.y-initialrightprint.y;
    double cs = cos ( initialrightprint.th );
    double ss = sin ( initialrightprint.th );
    sprConfig ( 0 ) = cs*drlx+ss*drly;
    sprConfig ( 1 ) = -ss*drlx+cs*drly;
    sprConfig ( 2 ) = drlth;

    cs = cos ( initialleftprint.th );
    ss = sin ( initialleftprint.th );
    splConfig ( 0 ) = -cs*drlx-ss*drly;
    splConfig ( 1 ) = ss*drlx-cs*drly;
    splConfig ( 2 ) = -drlth;

    bool currentRootIsRight = rootIsRight;
    for ( unsigned int i=0;i<numberExtraVirtualStepLinks;i++ )
    {
        if ( currentRootIsRight )
        {
            ChppLosSupportPolygonLink* spLink = new ChppLosSupportPolygonLink ( true, sprConfig ( 0 ),sprConfig ( 1 ),sprConfig ( 2 ), leftStepArea, righttStepArea ,inLeftFootShape, inRightFootShape );
            addLink ( spLink, ndof, config, outMask,outCentipedeParts,outCentipede,&inStandingRobot );
        }
        else
        {
            ChppLosSupportPolygonLink* spLink = new ChppLosSupportPolygonLink ( false, splConfig ( 0 ),splConfig ( 1 ),splConfig ( 2 ), leftStepArea, righttStepArea ,inLeftFootShape, inRightFootShape );
            addLink ( spLink, ndof, config, outMask,outCentipedeParts,outCentipede,&inStandingRobot );
        }
        currentRootIsRight = !currentRootIsRight;
    }

    ChppLosRobotLink* Rlink = new ChppLosRobotLink ( inStandingRobot, currentRootIsRight, leftStepArea, righttStepArea,inLeftFootShape, inRightFootShape );
    addLink ( Rlink, ndof, config, outMask, outCentipedeParts,outCentipede,&inStandingRobot );

    outCentipede->applyConfiguration ( config );

    return outCentipede;
}

void ChppLosCentipedeFactory::createCentipedeConstraints ( ChppLosCentipede* inCentipede, std::list<ChppLosTask*>& outCentipedeConstraints )
{
    for ( unsigned int i = 0;i <inCentipede->numberLinks();i++ )
    {
        ChppLosLink* ll;
        ll = inCentipede->linkAtRank ( i );
        outCentipedeConstraints.push_back ( new ChppLosFeetAntiCollisionTask ( *inCentipede, *ll ) ) ;
        outCentipedeConstraints.push_back ( new ChppLosFeetAreaTask ( *inCentipede,*ll ) );

        ChppLosRobotLink* rl = dynamic_cast<ChppLosRobotLink*> ( ll );
        if ( rl )
        {
            outCentipedeConstraints.push_back ( new ChppLosComBoundingTask ( *inCentipede,*rl ) );
            outCentipedeConstraints.push_back ( new ChppLosFeetProjectionTask ( *inCentipede,*rl ) );
        }

    }

}


void ChppLosCentipedeFactory::extendSpringman ( ChppLosCentipede& inCentipede, unsigned int extraSteps, std::list<ChppLosLink*>& outCentipedeParts, std::list<ChppLosTask*>& outCentipedeConstraints )
{
    if ( extraSteps == 0 )
        return;

    vectorN config;
    vectorN sprConfig ( 3 );
    vectorN splConfig ( 3 ),*spconfig;

    double drlth = inCentipede.initialLeftFootprint().th-inCentipede.initialRightFootprint().th;
    double drlx = inCentipede.initialLeftFootprint().x-inCentipede.initialRightFootprint().x;
    double drly = inCentipede.initialLeftFootprint().y-inCentipede.initialRightFootprint().y;
    double cs = cos ( inCentipede.initialRightFootprint().th );
    double ss = sin ( inCentipede.initialRightFootprint().th );
    sprConfig ( 0 ) = cs*drlx+ss*drly;
    sprConfig ( 1 ) = -ss*drlx+cs*drly;
    sprConfig ( 2 ) = drlth;

    cs = cos ( inCentipede.initialLeftFootprint().th );
    ss = sin ( inCentipede.initialLeftFootprint().th );
    splConfig ( 0 ) = -cs*drlx-ss*drly;
    splConfig ( 1 ) = ss*drlx-cs*drly;
    splConfig ( 2 ) = -drlth;

    ChppLosLink* firstLink = inCentipede.linkAtRank ( 0 );
    bool curRisRight = ! ( firstLink->isRootprintRight() );

    for ( unsigned int i=0;i<extraSteps;i++ )
    {
        if ( curRisRight )
            spconfig =& sprConfig;
        else
            spconfig =& splConfig;

        ChppLosSupportPolygonLink* spLink = new ChppLosSupportPolygonLink ( curRisRight, ( *spconfig ) ( 0 ), ( *spconfig ) ( 1 ), ( *spconfig ) ( 2 ),  firstLink->leftStepArea(), firstLink->rightStepArea(), firstLink->leftShapeArea(),  firstLink->rightShapeArea() );

        inCentipede.push_front ( *spLink );
        outCentipedeParts.push_back ( spLink );
        outCentipedeConstraints.push_front ( new ChppLosFeetAntiCollisionTask ( inCentipede,*spLink ) );
        outCentipedeConstraints.push_front ( new ChppLosFeetAreaTask ( inCentipede,*spLink ) );

        curRisRight = !curRisRight;

    }

    config.resize ( inCentipede.numberDof() );
    inCentipede.getConfiguration ( config );
    inCentipede.applyConfiguration ( config );


}



void ChppLosCentipedeFactory::addLink ( ChppLosLink* inLink, unsigned int& ndof, vectorN& config, vectorN& mask,  std::list<ChppLosLink*>& centipedeParts, ChppLosCentipede* centipede, ChppGikStandingRobot* inRobot )
{
    centipedeParts.push_back ( inLink );
    centipede->push_back ( *inLink );
    ndof += inLink->numberDof();
    config.resize ( ndof,true );
    subrange ( config,inLink->rank(), ndof ) = inLink->currentConfiguration();
    mask.resize ( ndof,true );
    if ( inLink->numberDof() == 3 )
        subrange ( mask,inLink->rank(), ndof ) = scalar_vector<double> ( 3,1 );
    else
        subrange ( mask,inLink->rank(), ndof ) = subrange ( inRobot->maskFactory()->wholeBodyMask(),6,inRobot->robot()->numberDof() );
}
