/*
  Copyright 2009 CNRS-LAAS

  Author: Oussama Kanoun
*/

#include "hpplocalstepper/tasks/HRP2BasicCollisionAvoider.hh"
#include <string.h>

#define M4_IJ MAL_S4x4_MATRIX_ACCESS_I_J

HRP2BasicCollisionAvoider::HRP2BasicCollisionAvoider ( ChppLosCentipede& inCentipede, ChppLosRobotLink& inLink ) :ChppLosTask ( inCentipede )
{
    attLink = &inLink;
    robot = attLink->standingRobot().robot();

    attConstraint->vector().resize ( 2 );
    attConstraint->matrix().resize ( 2, attCentipede->numberDof() );
    attConstraint->matrix().clear();
    attValue.resize ( 2 );
    
    noutM.resize ( 1,3 );

    tempJac3d.resize ( 3, robot->numberDof()-6 );
    tempJac3d.clear();
    jac1d.resize ( 1, robot->numberDof()-6 );
    jac1d.clear();

    attXi = 0.02;
    attDi = 0.05;
    attDs = 0.02;

    attWaist = robot->waist();
    attChest = robot->gazeJoint()->parentJoint()->parentJoint();
    
    locChest0 = vector3d(0.1,0.1,0.03);
    locChest1 = vector3d(0.1,-0.1,0.03);
    attChestzll = 0.31;
    
 
    attCA = new ChppLosCollisionAvoider(inCentipede,inLink );
    LosRobotCylinder cyl;
    
    LosRobotSphere sph;
    
    cyl.link = &inLink;
    cyl.joint = robot->rightWrist();
    cyl.locAB = vector3d(0,0,-0.14);
    cyl.locA = vector3d(0,0,0);
    cyl.locB = vector3d(0,0,-0.14);
    cyl.normAB=0.14;
    cyl.radius=0.07;
    attRobotCylinders.push_back(cyl);

   
    
    
    cyl.joint = robot->rightWrist()->parentJoint()->parentJoint();

    cyl.locAB = vector3d(0,0,-0.25);
    cyl.locA = vector3d(0,0,0);
    cyl.locB = vector3d(0,0,-0.25);
    cyl.normAB=0.25;
    cyl.radius=0.07;
    attRobotCylinders.push_back(cyl);

    
    
    cyl.joint = cyl.joint->parentJoint();
    cyl.locAB = vector3d(0,0,-0.25);
    cyl.locA = vector3d(0,0,0);
    cyl.locB = vector3d(0,0,-0.25);
    cyl.normAB=0.25;
    cyl.radius=0.07;
    attRobotCylinders.push_back(cyl);

    
    
    
    cyl.joint = robot->leftWrist();
    cyl.locAB = vector3d(0,0,-0.14);
    cyl.locA = vector3d(0,0,0);
    cyl.locB = vector3d(0,0,-0.14);
    cyl.normAB=0.14;
    cyl.radius=0.07;
    attRobotCylinders.push_back(cyl);
    
    
    cyl.joint = robot->leftWrist()->parentJoint()->parentJoint();
    cyl.locAB = vector3d(0,0,-0.25);
    cyl.locA = vector3d(0,0,0);
    cyl.locB = vector3d(0,0,-0.25);
    cyl.normAB=0.25;
    cyl.radius=0.07;
    attRobotCylinders.push_back(cyl);

    
    
    cyl.joint = cyl.joint->parentJoint();
    cyl.locAB = vector3d(0,0,-0.25);
    cyl.locA = vector3d(0,0,0);
    cyl.locB = vector3d(0,0,-0.25);
    cyl.normAB=0.25;
    cyl.radius=0.07;
    attRobotCylinders.push_back(cyl);

    
    
    cyl.joint = robot->waist();
    cyl.locAB = vector3d(0,0,0.23);
    cyl.locA = vector3d(-0.03,0,0.06);
    cyl.locB = vector3d(-0.03,0,0.29);
    cyl.normAB=0.23;
    cyl.radius=0.198;
    attRobotCylinders.push_back(cyl);

   

    cyl.joint =  robot->leftWrist()->parentJoint()->parentJoint()->parentJoint()->parentJoint()->parentJoint()->parentJoint();
    cyl.locAB = vector3d(0,0,0.25);
    cyl.locA = vector3d(-0.03,0,0.0);
    cyl.locB = vector3d(-0.03,0,0.25);
    cyl.normAB=0.25;
    cyl.radius=0.199;
    attRobotCylinders.push_back(cyl);

    
    
    
    cyl.joint = robot->rightFoot()->parentJoint()->parentJoint();
    cyl.locAB = vector3d(0,0,-0.34);
    cyl.locA = vector3d(0.0,-0.06,0.04);
    cyl.locB = vector3d(0.0,-0.06,-0.3);
    cyl.normAB=0.34;
    cyl.radius=0.11;
    attRobotCylinders.push_back(cyl);

    
    
    cyl.joint = cyl.joint->parentJoint();
    cyl.locAB = vector3d(0,0,-0.30);
    cyl.locA = vector3d(0.0,-0.06,0.0);
    cyl.locB = vector3d(0.0,-0.06,-0.3);
    cyl.normAB=0.30;
    cyl.radius=0.09;
    attRobotCylinders.push_back(cyl);

    
    
    cyl.joint = robot->leftFoot()->parentJoint()->parentJoint();
    cyl.locAB = vector3d(0,0,-0.34);
    cyl.locA = vector3d(0.0,0.06,0.04);
    cyl.locB = vector3d(0.0,0.06,-0.3);
    cyl.normAB=0.34;
    cyl.radius=0.11;
    attRobotCylinders.push_back(cyl);
   
    
    
    cyl.joint = cyl.joint->parentJoint();
    cyl.locAB = vector3d(0,0,-0.30);
    cyl.locA = vector3d(0.0,0.06,0.0);
    cyl.locB = vector3d(0.0,0.06,-0.3);
    cyl.normAB=0.30;
    cyl.radius=0.09;
    attRobotCylinders.push_back(cyl);
  
    
    
    sph.joint = robot->gazeJoint();
    sph.locS = vector3d(0.0,0.0,0.06);
    sph.radius=0.165;
    sph.link = &inLink;
    attRobotSpheres.push_back(sph);
  
    

    
    
    cyl.joint = robot->rightWrist();
    cyl.locAB = vector3d(0,0.03,0);
    cyl.locA = vector3d(0,-0.02,-0.14);
    cyl.locB = vector3d(0,0.01,-0.14);
    cyl.normAB=0.03;
    cyl.radius=0.08;
    attRobotCylinders.push_back(cyl);
    
    
    cyl.joint = robot->leftWrist();
    cyl.locAB = vector3d(0,0.03,0);
    cyl.locA = vector3d(0,-0.01,-0.14);
    cyl.locB = vector3d(0,0.02,-0.14);
    cyl.normAB=0.03;
    cyl.radius=0.08;
    attRobotCylinders.push_back(cyl);
    
#if 0
    unsigned int Rwrist = attCA->addRobotCylinder(&(attRobotCylinders[0]));
    unsigned int Rforearm = attCA->addRobotCylinder(&(attRobotCylinders[1]));
    unsigned int Rupperarm = attCA->addRobotCylinder(&(attRobotCylinders[2]));
    unsigned int Lwrist = attCA->addRobotCylinder(&(attRobotCylinders[3]));
    unsigned int Lforearm = attCA->addRobotCylinder(&(attRobotCylinders[4]));
    unsigned int Lupperarm = attCA->addRobotCylinder(&(attRobotCylinders[5]));
    unsigned int Waist = attCA->addRobotCylinder(&(attRobotCylinders[6]));
    unsigned int Chest = attCA->addRobotCylinder(&(attRobotCylinders[7]));
    unsigned int RTibia = attCA->addRobotCylinder(&(attRobotCylinders[8]));
    unsigned int Rthigh = attCA->addRobotCylinder(&(attRobotCylinders[9]));
    unsigned int LTibia = attCA->addRobotCylinder(&(attRobotCylinders[10]));
    unsigned int Lthigh = attCA->addRobotCylinder(&(attRobotCylinders[11]));
    unsigned int Head = attCA->addRobotSphere(&(attRobotSpheres[0]));
    unsigned int ROpenHand = attCA->addRobotCylinder(&(attRobotCylinders[12]));
    unsigned int LOpenHand = attCA->addRobotCylinder(&(attRobotCylinders[13]));
#endif    
}

HRP2BasicCollisionAvoider::~HRP2BasicCollisionAvoider()
{
    delete attCA;
}

void HRP2BasicCollisionAvoider::setMargins ( double inDi, double inDs, double inXi )
{
    attXi = inXi;
    attDs = inDs;
    attDi = inDi;
}


void HRP2BasicCollisionAvoider::computeConstraint()
{
    resize();

    waist[0] = M4_IJ ( attWaist->currentTransformation(),0,3 );
    waist[1] = M4_IJ ( attWaist->currentTransformation(),1,3 );
    waist[2] = M4_IJ ( attWaist->currentTransformation(),2,3 );

    calcChestConstraint(locChest0,0);
    
    calcChestConstraint(locChest1,1);

}

const vectorN& HRP2BasicCollisionAvoider::currentValue()
{
    attValue.clear();

    waist[0] = M4_IJ ( attWaist->currentTransformation(),0,3 );
    waist[1] = M4_IJ ( attWaist->currentTransformation(),1,3 );
    waist[2] = M4_IJ ( attWaist->currentTransformation(),2,3 );

    m = attChest->currentTransformation();
    chest0 = m*locChest0;
    chest1 = m*locChest1;

    calcChestValue(locChest0,0);
    
    calcChestValue(locChest1,1);
    
    return attValue;

}




void HRP2BasicCollisionAvoider::calcChestConstraint(const vector3d& inlocchest,unsigned int rank)
{
    ntonext[0] = M4_IJ ( attWaist->currentTransformation(),0,2 );
    ntonext[1] = M4_IJ ( attWaist->currentTransformation(),1,2 );
    ntonext[2] = M4_IJ ( attWaist->currentTransformation(),2,2 );
    
    m = attChest->currentTransformation();
    chest = m*inlocchest;

    otop = chest - waist;

    double value =  otop[0] *ntonext[0] + otop[1] *ntonext[1] + otop[2] *ntonext[2];

    unsigned int startRank = attLink->rank();

    robot->getPositionJacobian ( *attWaist, *attChest, inlocchest, tempJac3d, 0, false );

    noutM ( 0,0 ) = -ntonext[0];
    noutM ( 0,1 ) = -ntonext[1];
    noutM ( 0,2 ) = -ntonext[2];

    noalias ( jac1d ) = prod ( noutM,tempJac3d );
    memcpy ( ( &attConstraint->matrix().data() [rank*attCentipede->numberDof() +startRank] ),&jac1d.data() [0],attLink->numberDof() *sizeof ( double ) );

    if ( value < attChestzll + attDi )
        attConstraint->vector() ( rank ) = attXi* ( value - attChestzll - attDs ) / ( attDi - attDs );
    else
        attConstraint->vector() ( rank ) = attXi;
}

void HRP2BasicCollisionAvoider::calcChestValue(const vector3d& inlocchest,unsigned int rank)
{
    ntonext[0] = M4_IJ ( attWaist->currentTransformation(),0,2 );
    ntonext[1] = M4_IJ ( attWaist->currentTransformation(),1,2 );
    ntonext[2] = M4_IJ ( attWaist->currentTransformation(),2,2 );
    
    m = attChest->currentTransformation();
    chest = m*inlocchest;

    otop = chest - waist;

    double value =  otop[0] *ntonext[0] + otop[1] *ntonext[1] + otop[2] *ntonext[2];
    
    if ( value < attChestzll + attDs )
        attValue ( rank ) = value - attChestzll - attDs;
    else
        attValue ( rank ) = 0;
}


