/*
  Copyright 2009 CNRS-LAAS

  Author: Oussama Kanoun
*/

#include "boost/numeric/ublas/vector_proxy.hpp"
#include "boost/numeric/ublas/matrix_proxy.hpp"
#include "hpplocalstepper/tasks/hppLosConeTask.hh"

#define M3_IJ MAL_S3x3_MATRIX_ACCESS_I_J
#define M4_IJ MAL_S4x4_MATRIX_ACCESS_I_J

using namespace boost::numeric::ublas;

void ChppLosConeTask::equivAsymMat3 ( const vector3d& inVec, matrixNxP& outMat )
{
    outMat ( 0,0 ) = 0;
    outMat ( 0,1 ) = -inVec ( 2 );
    outMat ( 0,2 ) = inVec ( 1 );
    outMat ( 1,0 ) = inVec ( 2 );
    outMat ( 1,1 ) = 0;
    outMat ( 1,2 ) = -inVec ( 0 );
    outMat ( 2,0 ) = -inVec ( 1 );
    outMat ( 2,1 ) = inVec ( 0 );
    outMat ( 2,2 ) = 0;
}

ChppLosConeTask::ChppLosConeTask ( ChppLosCentipede& inCentipede, ChppLosRobotLink& inRobotLink, CjrlJoint& inJoint,const vector3d& inLocalVector,const vector3d& inWorldVector, double inMinCosine) :ChppLosTaskOnJoint ( inCentipede,inJoint )
{
    attRobotLink = &inRobotLink;
    attLocalVector = inLocalVector;
    attWorldVector = inWorldVector;
    attMinCosine = inMinCosine;
    attConstraint->matrix().resize ( 1,attCentipede->numberDof() );
    attConstraint->matrix().clear();
    attConstraint->vector().resize ( 1 );
    attTempJacobian.resize ( 3, attCentipede->numberDof() );
    attValue.resize ( 3 );
    attPlaneNormalM.resize ( 1,3 );
    attPlaneNormalM ( 0,0 ) = inWorldVector[0];
    attPlaneNormalM ( 0,1 ) = inWorldVector[1];
    attPlaneNormalM ( 0,2 ) = inWorldVector[2];
    attMaxVelocity = 1e10;
    
    attXi = 0.02;
    attDs = 0.02;
    attDi = 0.05;
}
bool ChppLosConeTask::resize()
{
    if ( ChppLosTask::resize() )
    {
        attTempJacobian.resize ( 3,attCentipede->numberDof() );
        return true;
    }
    return false;
}

void ChppLosConeTask::minCosine ( double inVal )
{
    attMinCosine = inVal;
}
            
const vector3d& ChppLosConeTask::localVector()
{
    return attLocalVector;
}
void ChppLosConeTask::localVector( const vector3d& inVec )
{
    attLocalVector= inVec;
}

const vector3d& ChppLosConeTask::worldVector()
{
    return attWorldVector;
}

void ChppLosConeTask::worldVector( const vector3d& inVec )
{
    attWorldVector = inVec;
}

void ChppLosConeTask::maxVelocityNorm ( double inMV )
{
    attMaxVelocity = ( inMV>0.0 ) ?inMV:0.0;
}

const vectorN& ChppLosConeTask::currentValue()
{
    for ( unsigned int i=0;i<3;i++ )
        for ( unsigned int j=0;j<3;j++ )
            M3_IJ(m3, i, j ) = M4_IJ(attJoint->currentTransformation(), i, j );
    attCurrentVector = m3*attLocalVector;
    attValue ( 0 ) = attCurrentVector[0]*attWorldVector[0]+attCurrentVector[1]*attWorldVector[1]+attCurrentVector[2]*attWorldVector[2]-attMinCosine;
    
    if (attValue(0) > 0)
       attValue(0) = 0;
    
    return attValue;
}

void ChppLosConeTask::computeConstraint()
{
    resize();
    
    for ( unsigned int i=0;i<3;i++ )
        for ( unsigned int j=0;j<3;j++ )
            M3_IJ(m3, i, j ) = M4_IJ(attJoint->currentTransformation(), i, j );
    attCurrentVector = m3*attLocalVector;

    
    attConstraint->vector() ( 0 ) = attCurrentVector[0]*attWorldVector[0]+attCurrentVector[1]*attWorldVector[1]+attCurrentVector[2]*attWorldVector[2]-attMinCosine;

    double nor = fabs ( attConstraint->vector() ( 0 ) );
    if ( nor > attMaxVelocity )
        attConstraint->vector() = attMaxVelocity*attConstraint->vector() /nor;

    attTempJacobian.clear();
    attRobotLink->standingRobot().robot()->getOrientationJacobian ( *attRobotLink->rootJoint(),*attJoint, attTempJacobian, attRobotLink->rank(), false );
    ChppLosLink* plink = attRobotLink->parent();
    while ( plink!=0 )
    {
        plink->getOrientationJacobian (attTempJacobian,plink->rank() );
        plink = plink->parent();
    }
    
    matrixNxP matVec(3,3), matProd(1,3);
    equivAsymMat3 ( attCurrentVector, matVec);
    noalias(matProd) = prod(attPlaneNormalM,matVec);
    noalias ( attConstraint->matrix() ) = prod ( matProd,attTempJacobian );
}

