/*
  Copyright 2009 CNRS-LAAS

  Author: Oussama Kanoun
*/

#include "hpplocalstepper/tasks/hppLosDiscAvoider.hh"
#include <string.h>

#define M4_IJ MAL_S4x4_MATRIX_ACCESS_I_J

ChppLosDiscAvoider::ChppLosDiscAvoider ( ChppLosCentipede& inCentipede ) :ChppLosTask ( inCentipede )
{
    attConstraint->vector().resize ( inCentipede.numberLinks() );
    attConstraint->matrix().resize ( inCentipede.numberLinks(), attCentipede->numberDof() );
    attConstraint->matrix().clear();

    attValue.resize ( inCentipede.numberLinks() );
    noutM.resize ( 1,3 );

    tempJac3d.resize ( 3, attCentipede->numberDof() );
    tempJac3d.clear();
    jac1d.resize ( 1, attCentipede->numberDof() );
    jac1d.clear();

    attXi = 0.20;
    attDi = 0.6;
    attDs = 0.3;

    attPositionX = 0.0;
    attPositionY = 0.0;
}

bool ChppLosDiscAvoider::resize()
{
    if ( attCentipede->numberLinks() != attValue.size() )
    {
        tempJac3d.resize ( 3,attCentipede->numberDof() );
        jac1d.resize ( 1,attCentipede->numberDof() );
        attConstraint->vector().resize ( attCentipede->numberLinks() );
        attConstraint->matrix().resize ( attCentipede->numberLinks(), attCentipede->numberDof() );
        attConstraint->matrix().clear();
        attValue.resize ( attCentipede->numberLinks() );
        return true;
    }
    return false;
}

void ChppLosDiscAvoider::setMargins ( double inDi, double inDs, double inXi )
{
    attXi = inXi;
    attDs = inDs;
    attDi = inDi;
}

void ChppLosDiscAvoider::setPosition ( double inX, double inY )
{
    attPositionX = inX;
    attPositionY = inY;
}


void ChppLosDiscAvoider::projectObstacleOnFoot( ChppLosLink* inLink, vector3d& constrainedPoint)
{
    double cs = cos(inLink->frontprint().th);
    double ss = sin(inLink->frontprint().th);
    
    double rx = cs * ( attPositionX - inLink->frontprint().x) + ss * ( attPositionY - inLink->frontprint().y);
    double ry = cs * (attPositionY - inLink->frontprint().y) - ss * ( attPositionX - inLink->frontprint().x);
    
    double rpx,rpy;
    
    const ChppLosFootprintShape* shape; 
    if (inLink->rootIsRight())
        shape = &inLink->leftShapeArea();
    else
        shape = &inLink->rightShapeArea();
    
    if (ry > shape->yUpLeft)
        rpy = shape->yUpLeft;
    else if (ry < shape->yUpRight)
        rpy = shape->yUpRight;
    else 
        rpy = ry;

    if (rx > shape->xUpLeft)
        rpx = shape->xUpLeft;
    else if (rx > shape->xDownLeft)
        rpx = rx;
    else
        rpx = shape->xDownLeft;
    
    
    constrainedPoint[0] = cs * rpx - ss * rpy + inLink->frontprint().x;
    constrainedPoint[1] = ss * rpx + cs * rpy + inLink->frontprint().y;
    constrainedPoint[2] = 0;
}
    
void ChppLosDiscAvoider::calcRegionConstraint ( ChppLosLink* inLink, unsigned int rank )
{
    projectObstacleOnFoot(inLink,proj);
            
    otop[0] = attPositionX - proj[0];
    otop[1] = attPositionY - proj[1];
    otop[2] = 0.0;

    double dist = sqrt ( otop[0]*otop[0] + otop[1]*otop[1] );

    otop[0] /= dist;
    otop[1] /= dist;

    ChppLosLink* plink = inLink;
    tempJac3d.clear();
    while ( plink!=0 )
    {
        plink->getPositionJacobian ( proj, tempJac3d, plink->rank() );
        plink = plink->parent();
    }


    noutM ( 0,0 ) = otop[0];
    noutM ( 0,1 ) = otop[1];
    noutM ( 0,2 ) = otop[2];

    noalias ( jac1d ) = prod ( noutM,tempJac3d );
    row ( attConstraint->matrix(),rank ) = row ( jac1d,0 );

    if ( dist < attDi )
        attConstraint->vector() ( rank ) = attXi* ( dist - attDs ) / ( attDi - attDs );
    else
        attConstraint->vector() ( rank ) = attXi;
}

void ChppLosDiscAvoider::calcRegionValue ( ChppLosLink* inLink, unsigned int rank )
{
    double dist = sqrt ( pow ( inLink->frontprint().x - attPositionX,2 ) +  pow ( inLink->frontprint().y - attPositionY,2 ) );

    if ( dist < attDs )
        attValue ( rank ) = dist - attDs;
    else
        attValue ( rank ) = 0;
}


void ChppLosDiscAvoider::computeConstraint()
{
    resize();

    for (unsigned int i = 0; i<attCentipede->numberLinks();i++)
        calcRegionConstraint ( attCentipede->linkAtRank(i),  i );
}

const vectorN& ChppLosDiscAvoider::currentValue()
{
    attValue.clear();

    for (unsigned int i = 0; i<attCentipede->numberLinks();i++)
        calcRegionValue ( attCentipede->linkAtRank(i),  i );

    return attValue;

}

