/*
  Copyright 2009 CNRS-LAAS

  Author: Oussama Kanoun
*/

#include "hpplocalstepper/tasks/hppLosFeetProjectionTask.hh"
#include "hppGik/hppGikTools.h"

#define M3_IJ MAL_S3x3_MATRIX_ACCESS_I_J
#define M4_IJ MAL_S4x4_MATRIX_ACCESS_I_J

void ChppLosFeetProjectionTask::equivAsymMat3 ( const vector3d& inVec, matrixNxP& outMat )
{
    outMat ( 0,0 ) = 0;
    outMat ( 0,1 ) = -inVec ( 2 );
    outMat ( 0,2 ) = inVec ( 1 );
    outMat ( 1,0 ) = inVec ( 2 );
    outMat ( 1,1 ) = 0;
    outMat ( 1,2 ) = -inVec ( 0 );
    outMat ( 2,0 ) = -inVec ( 1 );
    outMat ( 2,1 ) = inVec ( 0 );
    outMat ( 2,2 ) = 0;
}

void ChppLosFeetProjectionTask::crossProd3 ( const vector3d& v1, const vector3d& v2, vector3d& retV )
{
    retV ( 0 ) = v1 ( 1 ) *v2 ( 2 ) - v1 ( 2 ) *v2 ( 1 );
    retV ( 1 ) = v1 ( 2 ) *v2 ( 0 ) - v1 ( 0 ) *v2 ( 2 );
    retV ( 2 ) = v1 ( 0 ) *v2 ( 1 ) - v1 ( 1 ) *v2 ( 0 );
}



ChppLosFeetProjectionTask::ChppLosFeetProjectionTask ( ChppLosCentipede& inCentipede, ChppLosRobotLink& inRobotLink ) :ChppLosTask ( inCentipede )
{
    attRobotLink = &inRobotLink;
    attStandingRobot = &inRobotLink.standingRobot();

    attConstraint->matrix().resize ( 4,attCentipede->numberDof() );
    attConstraint->matrix().clear();
    attConstraint->vector().resize ( 4 );
    jacF.resize ( 6,attRobotLink->numberDof() );
    tempRot.resize ( 3,3 );
    tempPos.resize ( 3 );
    tempOmega = tempPos;
    attValue.resize ( 2 );
    matOP.resize ( 3,3 );
    matOT.resize ( 3,3 );
    matOTOP.resize ( 3,3 );
    MAL_S3x3_MATRIX_SET_IDENTITY(m3);
}



void ChppLosFeetProjectionTask::computeConstraint()
{
    resize();
    //Foot Slide constraints
    CjrlHumanoidDynamicRobot* jrlRobot = attStandingRobot->robot();
    unsigned int startRank = attRobotLink->rank();

    
    for ( unsigned int i=0;i<3;i++ ) for ( unsigned int j=0;j<3;j++ ) M3_IJ ( m3, i, j ) = M4_IJ ( attRobotLink->frontJoint()->currentTransformation(), i, j );
    for ( unsigned int i=0;i<3;i++ ) locVec[i] = M3_IJ ( m3, i, 2 );
    
    vector3d cross;
    crossProd3 ( vector3d(0,0,1),locVec,cross );
    attConstraint->vector() ( 1 ) = cross ( 0 );
    attConstraint->vector() ( 2 ) = cross ( 1 );
    attConstraint->vector() ( 3 ) = cross ( 2 );
    attConstraint->vector() (0) =  attStandingRobot->robot()->footHeight()-M4_IJ(attRobotLink->frontJoint()->currentTransformation(), 2, 3 );
    
    jrlRobot->getJacobian ( *attRobotLink->rootJoint(),*attRobotLink->frontJoint(), vector3d ( 0,0,0 ), jacF,0,false );
    
    subrange ( attConstraint->matrix(),0,1,startRank,startRank+attRobotLink->numberDof() ) =
            subrange ( jacF,2,3,0,attRobotLink->numberDof() );
    
    equivAsymMat3 ( vector3d(0,0,1),matOT );
    equivAsymMat3 ( locVec,matOP );
    noalias ( matOTOP ) = prod ( matOT,matOP );
    subrange ( attConstraint->matrix(),1,4,startRank, startRank + attRobotLink->numberDof() ) = prod ( matOTOP,subrange(jacF,3,6,0, attRobotLink->numberDof()));
    
    
    
}


const vectorN& ChppLosFeetProjectionTask::currentValue()
{
    attValue ( 0 ) = attStandingRobot->robot()->footHeight()-M4_IJ(attRobotLink->frontJoint()->currentTransformation(), 2, 3 );//dz

    for ( unsigned int i=0;i<3;i++ )for ( unsigned int j=0;j<3;j++ ) M3_IJ(m3, i, j ) = M4_IJ(attRobotLink->frontJoint()->currentTransformation(), i, j );
    for ( unsigned int i=0;i<3;i++ ) locVec[i] = M3_IJ ( m3, i, 2 );
    crossProd3 ( vector3d(0,0,1),locVec,cross );
    double sv = MAL_S3_VECTOR_NORM ( cross );
    double cv = locVec ( 2 );
    attValue ( 1 ) = atan2 ( sv,cv );
    
    return attValue;
}
