/*
  Copyright 2009 CNRS-LAAS

  Author: Oussama Kanoun
*/

#include "boost/numeric/ublas/vector_proxy.hpp"
#include "boost/numeric/ublas/matrix_proxy.hpp"
#include "hpplocalstepper/tasks/hppLosPlaneTask.hh"

using namespace boost::numeric::ublas;

ChppLosPlaneTask::ChppLosPlaneTask ( ChppLosCentipede& inCentipede, ChppLosRobotLink& inRobotLink, CjrlJoint& inJoint,const vector3d& inLocalPoint,const vector3d& inTargetPoint , const vector3d& inTargetNormal ) :ChppLosTaskOnJoint ( inCentipede,inJoint )
{
    attRobotLink = &inRobotLink;
    attLocalPoint = inLocalPoint;
    attPlanePoint = inTargetPoint;
    attPlaneNormal = inTargetNormal;

    attConstraint->matrix().resize ( 1,attCentipede->numberDof() );
    attConstraint->matrix().clear();
    attConstraint->vector().resize ( 1 );
    attTempJacobian.resize ( 3, attCentipede->numberDof() );
    attValue.resize ( 3 );
    attPlaneNormalM.resize ( 1,3 );
    attPlaneNormalM ( 0,0 ) = attPlaneNormal[0];
    attPlaneNormalM ( 0,1 ) = attPlaneNormal[1];
    attPlaneNormalM ( 0,2 ) = attPlaneNormal[2];
    attMaxVelocity = 1e10;
    
    attIsEquality = true;
}
bool ChppLosPlaneTask::resize()
{
    if ( ChppLosTask::resize() )
    {
        attTempJacobian.resize ( 3,attCentipede->numberDof() );
        return true;
    }
    return false;
}

const vector3d& ChppLosPlaneTask::localPoint()
{
    return attLocalPoint;
}
void ChppLosPlaneTask::localPoint ( const vector3d& inVec )
{
    attLocalPoint = inVec;
}

const vector3d& ChppLosPlaneTask::targetPlanePoint()
{
    return attPlanePoint;
}

const vector3d& ChppLosPlaneTask::targetPlaneNormal()
{
    return attPlaneNormal;
}

void ChppLosPlaneTask::targetPlanePoint ( const vector3d& inVec )
{
    attPlanePoint = inVec;
}

void ChppLosPlaneTask::targetPlaneNormal ( const vector3d& inVec )
{
    attPlaneNormal = inVec;
    attPlaneNormalM ( 0,0 ) = attPlaneNormal[0];
    attPlaneNormalM ( 0,1 ) = attPlaneNormal[1];
    attPlaneNormalM ( 0,2 ) = attPlaneNormal[2];
}

void ChppLosPlaneTask::maxVelocityNorm ( double inMV )
{
    attMaxVelocity = ( inMV>0.0 ) ?inMV:0.0;
}

const vectorN& ChppLosPlaneTask::currentValue()
{
    m = attJoint->currentTransformation();
    attCurrentPoint = m * attLocalPoint;
    tempV = attPlanePoint-attCurrentPoint;
    attValue ( 0 ) = tempV[0]*attPlaneNormal[0]+tempV[1]*attPlaneNormal[1]+tempV[2]*attPlaneNormal[2];
    //std::cout << attPlanePoint<< " "<< attCurrentPoint <<" "<<attValue ( 0 ) << std::endl;
    if (!attIsEquality)
        if (attValue(0) > 0)
            attValue(0) = 0;
    return attValue;
}

void ChppLosPlaneTask::computeConstraint()
{
    resize();
    m = attJoint->currentTransformation();
    attCurrentPoint = m * attLocalPoint;
    tempV = attPlanePoint - attCurrentPoint;
    attConstraint->vector() ( 0 ) = tempV[0]*attPlaneNormal[0]+tempV[1]*attPlaneNormal[1]+tempV[2]*attPlaneNormal[2];

    double nor = fabs ( attConstraint->vector() ( 0 ) );
    if ( nor > attMaxVelocity )
        attConstraint->vector() = attMaxVelocity*attConstraint->vector() /nor;

    attTempJacobian.clear();
    attRobotLink->standingRobot().robot()->getPositionJacobian ( *attRobotLink->rootJoint(),*attJoint, attLocalPoint, attTempJacobian, attRobotLink->rank(), false );
    ChppLosLink* plink = attRobotLink->parent();
    while ( plink!=0 )
    {
        plink->getPositionJacobian ( attCurrentPoint,attTempJacobian,plink->rank() );
        plink = plink->parent();
    }
    noalias ( attConstraint->matrix() ) = prod ( attPlaneNormalM,attTempJacobian );
}

