/*
  Copyright 2009 CNRS-LAAS

  Author: Oussama Kanoun
*/

#include "boost/numeric/ublas/vector_proxy.hpp"
#include "boost/numeric/ublas/matrix_proxy.hpp"
#include "hpplocalstepper/tasks/hppLosPositionTask.hh"

using namespace boost::numeric::ublas;

ChppLosPositionTask::ChppLosPositionTask (ChppLosCentipede& inCentipede, ChppLosRobotLink& inRobotLink, CjrlJoint& inJoint,const vector3d& inLocalPoint,const vector3d& inWorlTarget ):ChppLosTaskOnJoint(inCentipede,inJoint)
{
    attRobotLink = &inRobotLink;
    attLocalPoint = inLocalPoint;
    attWorldTarget = inWorlTarget;

    attConstraint->matrix().resize(3,attCentipede->numberDof());
    attConstraint->matrix().clear();
    attConstraint->vector().resize(3);
    attValue.resize(3);
    attMaxVelocity = 1e10;
}



const vector3d& ChppLosPositionTask::localPoint()
{
    return attLocalPoint;
}
void ChppLosPositionTask::localPoint ( const vector3d& inVec )
{
    attLocalPoint = inVec;
}

const vector3d& ChppLosPositionTask::targetPoint()
{
    return attWorldTarget;
}

void ChppLosPositionTask::targetPoint ( const vector3d& inVec )
{
    attWorldTarget = inVec;
}

void ChppLosPositionTask::maxVelocityNorm(double inMV)
{
    attMaxVelocity = (inMV>0.0)?inMV:0.0;
}

const vectorN& ChppLosPositionTask::currentValue()
{
    m =attJoint->currentTransformation();
    attCurrentPoint = m * attLocalPoint;
    attValue(0) = attWorldTarget[0] - attCurrentPoint[0];
    attValue(1) = attWorldTarget[1] - attCurrentPoint[1];
    attValue(2) = attWorldTarget[2] - attCurrentPoint[2];
    return attValue;
}

void ChppLosPositionTask::computeConstraint()
{
    resize();
    m =attJoint->currentTransformation();
    attCurrentPoint = m * attLocalPoint;
    attConstraint->vector()(0) = attWorldTarget[0] - attCurrentPoint[0];
    attConstraint->vector()(1) = attWorldTarget[1] - attCurrentPoint[1];
    attConstraint->vector()(2) = attWorldTarget[2] - attCurrentPoint[2];
    
    double nor = norm_2(attConstraint->vector());
    if (nor > attMaxVelocity)
        attConstraint->vector() = attMaxVelocity*attConstraint->vector()/nor;
    
    attConstraint->matrix().clear();
    attRobotLink->standingRobot().robot()->getPositionJacobian ( *attRobotLink->rootJoint(),*attJoint, attLocalPoint, attConstraint->matrix(), attRobotLink->rank(), false );
    ChppLosLink* plink = attRobotLink->parent();
    while ( plink!=0 )
    {
        plink->getPositionJacobian ( attCurrentPoint,attConstraint->matrix(),plink->rank() );
        plink = plink->parent();
    }
}

