/*
  Copyright 2009 CNRS-LAAS

  Author: Oussama Kanoun
*/

#include "hppLocalStepperTest.hh"

#include "hppGik/hppGikTools.h"
#include "hppGik/robot/hppRobotMotion.h"

#include "hpplocalstepper/tasks/hppLosComBoundingTask.hh"
#include "hpplocalstepper/tasks/hppLosComTask.hh"
#include "hpplocalstepper/tasks/hppLosFeetProjectionTask.hh"
#include "hpplocalstepper/tasks/hppLosFeetAntiCollisionTask.hh"
#include "hpplocalstepper/tasks/hppLosFeetAreaTask.hh"
#include "hpplocalstepper/tasks/hppLosDiscAvoider.hh"

#include "hpplocalstepper/tasks/hppLosPositionTask.hh"
#include "hpplocalstepper/tasks/hppLosParallelTask.hh"
#include "hpplocalstepper/tasks/hppLosPlaneTask.hh"
#include "hpplocalstepper/tasks/hppLosCoplanarTask.hh"
#include "hpplocalstepper/tasks/hppLosLinkConfigurationTask.hh"
#include "hpplocalstepper/tasks/hppLosGazeTask.hh"
#include "hpplocalstepper/hppLosSolver.hh"
#include "hpplocalstepper/hppLosCentipedeFactory.hh"
#include "hpplocalstepper/tasks/HRP2BasicCollisionAvoider.hh"
#include "hppLosMotionGenerator.hh"



#define M3_IJ MAL_S3x3_MATRIX_ACCESS_I_J
#define M4_IJ MAL_S4x4_MATRIX_ACCESS_I_J

using namespace boost::numeric::ublas;

ChppGikStandingRobot* ChppLocalStepperTest::createHumanoidRobot()
{
    CjrlRobotDynamicsObjectConstructor<
    dynamicsJRLJapan::DynamicMultiBody,
    //dynamicsJRLJapan::HumanoidDynamicMultiBody,
    Chrp2OptHumanoidDynamicRobot,
    dynamicsJRLJapan::JointFreeflyer,
    dynamicsJRLJapan::JointRotation,
    dynamicsJRLJapan::JointTranslation,
    dynamicsJRLJapan::Body> jrlRobotFactory;

    CjrlHumanoidDynamicRobot* attRobot = jrlRobotFactory.createhumanoidDynamicRobot();

    dynamicsJRLJapan::HumanoidDynamicMultiBody * aHDMB = ( dynamicsJRLJapan::HumanoidDynamicMultiBody* ) attRobot;

    std::string path = "./";
    std::string name = "HRP2.wrl";
    aHDMB->parserVRML ( path,name,"./HRP2LinkJointRank.xml" );
    std::string aName="./HRP2Specificities.xml";
    aHDMB->SetHumanoidSpecificitiesFile ( aName );
    //aHDMB->SetTimeStep ( attSamplingPeriod );
    aHDMB->setComputeCoM ( true );


    aHDMB->setComputeVelocity ( false );
    aHDMB->setComputeMomentum ( false );
    aHDMB->setComputeAcceleration ( false );
    aHDMB->setComputeZMP ( false );
    aHDMB->setComputeSkewCoM ( false );
    aHDMB->setComputeAccelerationCoM ( false );
    aHDMB->setComputeBackwardDynamics ( false );



    unsigned int nDof = attRobot->numberDof();
    vectorN halfsittingConf ( nDof );

    double rhandOP = 10;
    double lhandOP = 10;

    //Half sitting
    double dInitPos[40] =
    {
        0.0, 0.0, -26.0, 50.0, -24.0, 0.0, // right leg

        0.0, 0.0, -26.0, 50.0, -24.0, 0.0, // left leg

        0.0, 0.0, // chest

        0.0, 0.0, // head

        15.0, -10.0, 0.0, -30.0, 0.0, 0.0, // right arm

        rhandOP, // right hand clench

        15.0,  10.0, 0.0, -30.0, 0.0, 0.0, // left arm

        lhandOP, // left hand clench

        -rhandOP, rhandOP, -rhandOP, rhandOP, -rhandOP, // right hand parallel mechanism
        -lhandOP, lhandOP, -lhandOP, lhandOP, -lhandOP  // left hand parallel mechanism
    };

    //waist x y z
    halfsittingConf ( 0 ) = 1.0;
    halfsittingConf ( 1 ) = -0.8;
    halfsittingConf ( 2 ) = 0.6487;
    //waist roll pitch yaw
    halfsittingConf ( 3 ) = 0.0;
    halfsittingConf ( 4 ) = 0.0;
    halfsittingConf ( 5 ) = -3*M_PI/4;

    //joints
    for ( unsigned int i=6;i<nDof;i++ )
        halfsittingConf ( i ) = dInitPos[i-6]*M_PI/180;

    zero_vector<double> zeros ( attRobot->numberDof() );
    attRobot->currentConfiguration ( halfsittingConf );
    attRobot->currentVelocity ( zeros );
    attRobot->currentAcceleration ( zeros );
    attRobot->computeForwardKinematics();

    //set gaze origin and direction
    vector3d gazeDir,gazeOrigin;
    gazeDir[0] = 1;
    gazeDir[1] = 0;
    gazeDir[2] = 0;

    gazeOrigin[0] = 0;
    gazeOrigin[1] = 0;
    gazeOrigin[2] = 0.118;

    aHDMB->gaze ( ( const vector3d& ) gazeDir, ( const vector3d& ) gazeOrigin );

    ChppGik2DShape leftFootShape, rightFootShape;

    ChppGik2DVertex vert;

    vert.x = 0.135;
    vert.y = 0.079;
    leftFootShape.vertices.push_back ( vert );
    vert.x = -0.105;
    leftFootShape.vertices.push_back ( vert );
    vert.y = -0.059;
    leftFootShape.vertices.push_back ( vert );
    vert.x = 0.135;
    leftFootShape.vertices.push_back ( vert );

    vert.x = 0.135;
    vert.y = 0.059;
    rightFootShape.vertices.push_back ( vert );
    vert.x = -0.105;
    rightFootShape.vertices.push_back ( vert );
    vert.y = -0.079;
    rightFootShape.vertices.push_back ( vert );
    vert.x = 0.135;
    rightFootShape.vertices.push_back ( vert );

    ChppGikStandingRobot* standingRobot = new ChppGikStandingRobot ( *attRobot, leftFootShape, rightFootShape );


    standingRobot->staticState ( halfsittingConf );
    return standingRobot;
}

ChppLocalStepperTest::ChppLocalStepperTest() // : attSamplingPeriod ( 5e-3 )
{
    createCentipede();
}

void ChppLocalStepperTest::createCentipede()
{
    ChppGikStandingRobot* sr1 = createHumanoidRobot();
    standingRobots.push_back ( sr1 );
    ChppLosFootprint iLeft,iRight;
    matrix4d mf = sr1->robot()->rightFoot()->currentTransformation();
    iRight.x = M4_IJ ( mf,0,3 );
    iRight.y =  M4_IJ ( mf,1,3 );
    iRight.th =  atan2 ( M4_IJ ( mf,1,0 ),M4_IJ ( mf,0,0 ) );

    mf = sr1->robot()->leftFoot()->currentTransformation();
    iLeft.x = M4_IJ ( mf,0,3 );
    iLeft.y =  M4_IJ ( mf,1,3 );
    iLeft.th =  atan2 ( M4_IJ ( mf,1,0 ),M4_IJ ( mf,0,0 ) );

    ChppLosFootprintShape lshape;
    lshape.xDownLeft = -0.105;
    lshape.xDownRight = -0.105;
    lshape.xUpRight = 0.135;
    lshape.xUpLeft = 0.135;
    lshape.yDownLeft = 0.079;
    lshape.yUpLeft = 0.079;
    lshape.yDownRight = -0.059;
    lshape.yUpRight = -0.059;

    ChppLosFootprintArea larea ;
    larea.maxX = 0.22;
    larea.minX = -0.22;
    larea.maxY = 0.25;
    larea.minY = 0.07;
    larea.maxTheta = M_PI/4;
    larea.minTheta = -0.1;

    ChppLosFootprintShape rshape;
    rshape.xDownLeft = -0.105;
    rshape.xDownRight = -0.105;
    rshape.xUpRight = 0.135;
    rshape.xUpLeft = 0.135;
    rshape.yDownLeft = 0.059;
    rshape.yUpLeft = 0.059;
    rshape.yDownRight = -0.079;
    rshape.yUpRight = -0.079;

    ChppLosFootprintArea rarea ;
    rarea.maxX = larea.maxX;
    rarea.minX = larea.minX;
    rarea.maxY = -larea.minY;
    rarea.minY = -larea.maxY;
    rarea.maxTheta = -larea.minTheta;
    rarea.minTheta = -larea.maxTheta;

    unsigned int nSteps = 15;

    attCentipede = ChppLosCentipedeFactory::createSpringman ( iLeft,iRight, *sr1, nSteps, false, larea,  rarea, lshape, rshape, attCentipedeParts, attMask );
    centipedes.push_back ( attCentipede );

}


ChppLosCentipede* ChppLocalStepperTest::createAnotherCentipede ( ChppGikStandingRobot* sr1, unsigned int inExtraSteps, bool rootIsRight )
{

    ChppLosFootprint iLeft,iRight;
    matrix4d mf = sr1->robot()->rightFoot()->currentTransformation();
    iRight.x = M4_IJ ( mf,0,3 );
    iRight.y =  M4_IJ ( mf,1,3 );
    iRight.th =  atan2 ( M4_IJ ( mf,1,0 ),M4_IJ ( mf,0,0 ) );

    mf = sr1->robot()->leftFoot()->currentTransformation();
    iLeft.x = M4_IJ ( mf,0,3 );
    iLeft.y =  M4_IJ ( mf,1,3 );
    iLeft.th =  atan2 ( M4_IJ ( mf,1,0 ),M4_IJ ( mf,0,0 ) );

    ChppLosFootprintShape lshape;
    lshape.xDownLeft = -0.105;
    lshape.xDownRight = -0.105;
    lshape.xUpRight = 0.135;
    lshape.xUpLeft = 0.135;
    lshape.yDownLeft = 0.079;
    lshape.yUpLeft = 0.079;
    lshape.yDownRight = -0.059;
    lshape.yUpRight = -0.059;

    ChppLosFootprintArea larea ;
    larea.maxX = 0.22;
    larea.minX = -0.22;
    larea.maxY = 0.25;
    larea.minY = 0.07;
    larea.maxTheta = M_PI/4;
    larea.minTheta = -0.1;

    ChppLosFootprintShape rshape;
    rshape.xDownLeft = -0.105;
    rshape.xDownRight = -0.105;
    rshape.xUpRight = 0.135;
    rshape.xUpLeft = 0.135;
    rshape.yDownLeft = 0.059;
    rshape.yUpLeft = 0.059;
    rshape.yDownRight = -0.079;
    rshape.yUpRight = -0.079;

    ChppLosFootprintArea rarea ;
    rarea.maxX = larea.maxX;
    rarea.minX = larea.minX;
    rarea.maxY = -larea.minY;
    rarea.minY = -larea.maxY;
    rarea.maxTheta = -larea.minTheta;
    rarea.minTheta = -larea.maxTheta;


    return ChppLosCentipedeFactory::createSpringman ( iLeft,iRight, *sr1, inExtraSteps, rootIsRight, larea,  rarea, lshape, rshape, attCentipedeParts, attMask );

}

ChppLocalStepperTest::~ChppLocalStepperTest()
{
    for ( unsigned int i=0;i<standingRobots.size();i++ )
        delete standingRobots[i];

    for ( unsigned int i=0;i<centipedes.size();i++ )
        delete centipedes[i];

    for ( unsigned int i=0;i<attCentipedeParts.size();i++ )
    {
        delete * ( attCentipedeParts.begin() );
        attCentipedeParts.pop_front();
    }
}


void ChppLocalStepperTest::plan()
{
    
    std::vector<ChppLosRobotLink*> robotslinks;
    std::vector<CjrlHumanoidDynamicRobot*> robots;
    std::list<ChppLosRobotLink*>::iterator iterRobot;
    std::vector<ChppLosTask*> tasks,constraints;
    std::list<ChppLosTask*> centipedeConstraints;
    
    vectorN hsc = subrange ( standingRobots[0]->halfsittingConfiguration(),6,46 );
    vectorN configMask = subrange ( standingRobots[0]->maskFactory()->wholeBodyMask(),6,46 );
    
    for ( iterRobot = attCentipede->robotLinks().begin(); iterRobot!=attCentipede->robotLinks().end(); iterRobot++ )
    {
        robotslinks.push_back ( *iterRobot );
        robots.push_back ( ( *iterRobot )->standingRobot().robot() );
    }

    //Create solver
    ChppLosSolver solver1 ( *attCentipede );
    
    //Left Hand  position task
    ChppLosPositionTask handposTL ( *attCentipede,* robotslinks[0],* robots[0]->leftWrist() ,robots[0]->leftHand()->centerInWristFrame(), vector3d ( 0.3,-3.0,1.08 ) );
    handposTL.maxVelocityNorm ( 0.02 );
    handposTL.dampingFactor ( 0.2 );
    tasks.push_back ( &handposTL );

    //Right Hand  position task
    ChppLosPositionTask handposTR ( *attCentipede,* robotslinks[0],* robots[0]->rightWrist() ,robots[0]->rightHand()->centerInWristFrame(), vector3d ( 0.0,-3.0,1.08 ) );
    handposTR.maxVelocityNorm ( 0.02 );
    handposTR.dampingFactor ( 0.2 );
    tasks.push_back ( &handposTR );

    //Task on waist to remain horizontal
    ChppLosParallelTask parC ( *attCentipede, * robotslinks[0], *robots[0]->waist(),vector3d ( 0,0,1 ),vector3d ( 0,0,1 ) );
    parC.dampingFactor ( 0.2 );
    parC.maxVelocityNorm ( 0.2 );
    tasks.push_back ( &parC );

    //Task on configuration to remain at half-sitting
    ChppLosLinkConfigurationTask configT1 ( *attCentipede,* robotslinks[0], hsc,configMask );
    configT1.dampingFactor ( 0.2 );
    configT1.maxVelocityNorm ( 0.2 );
    tasks.push_back ( &configT1 );

    //Centipede constraints (feet sliding on ground, feet anticollision, com in support polygon)
    ChppLosCentipedeFactory::createCentipedeConstraints ( attCentipede, centipedeConstraints );
    std::list<ChppLosTask*>::iterator iter;
    for ( iter = centipedeConstraints.begin();iter != centipedeConstraints.end();iter++ )
        constraints.push_back ( *iter );

    //Create a disc on the ground to avoid
    ChppLosDiscAvoider discAvoider ( *attCentipede );
    discAvoider.setPosition ( -0.2,-1.4 );
    discAvoider.setMargins ( 1.1,0.9,0.2 );
    constraints.push_back ( &discAvoider );

    //Simplified collision avoidance constraint for hrp2
    HRP2BasicCollisionAvoider hrp2AvoiderC ( *attCentipede,*robotslinks[0] );
    constraints.push_back ( &hrp2AvoiderC );
    
    //a vertical plane to avoid
    LosExtPlane plane11;
    plane11.P = vector3d(0.0,-2.85,0.0);
    plane11.N = vector3d(0.0,1.0,0);
    plane11.ds = 0.00;
    
    //register the plane in the collision avoider
    unsigned int pi = hrp2AvoiderC.pairCollisionAvoider().addExtPlane(&plane11);
            
    //collect footprints to be checkecd for collision with plane
    std::vector<LosLinkFootprint> vecLinkFootprints;
    LosLinkFootprint lfp;
    for ( unsigned int i=0;i<attCentipede->numberLinks();i++ )
    {
        lfp.link = attCentipede->linkAtRank ( i );
        lfp.ds = 0.0;
        vecLinkFootprints.push_back ( lfp );
    }
    
    
    for(unsigned int i=0;i<vecLinkFootprints.size();i++)
    {
        //register the footprint in the collision avoider
        unsigned int  index = hrp2AvoiderC.pairCollisionAvoider().addLinkFootprint(&vecLinkFootprints[i]);
        //make pair to be checked in collision avoider
        hrp2AvoiderC.pairCollisionAvoider().makeLinkFootprintPlanePair ( index,pi );
    }
    
    //make pairs for collision checking (self-collision)
    hrp2AvoiderC.pairCollisionAvoider().makeRobotCylinderPair ( 4,6 );
    hrp2AvoiderC.pairCollisionAvoider().makeRobotCylinderPair ( 3,6 );
    hrp2AvoiderC.pairCollisionAvoider().makeRobotCylinderPair ( 4,7 );
    hrp2AvoiderC.pairCollisionAvoider().makeRobotCylinderPair ( 3,7 );
    hrp2AvoiderC.pairCollisionAvoider().makeRobotCylinderPair ( 4,0 );
    hrp2AvoiderC.pairCollisionAvoider().makeRobotCylinderPair ( 3,0 );
    hrp2AvoiderC.pairCollisionAvoider().setMargins(0.03,0.0,0.02);
    constraints.push_back ( & ( hrp2AvoiderC.pairCollisionAvoider() ) );

    
    solver1.constraints ( constraints );
    solver1.tasks ( tasks );
    solver1.setMask ( attMask );
    solver1.solve ( true,"motion" );
    std::cout << "Dumped iterations to files motion-footprints and motion-configs" << std::endl;

    for ( unsigned int i = 0;i <centipedeConstraints.size();i++ )
    {
        delete * ( centipedeConstraints.begin() );
        centipedeConstraints.pop_front();
    }
}


