/*
  Copyright 2009 CNRS-LAAS

  Author: Oussama Kanoun
*/

#include "hppLosMotionGenerator.hh"

#include "hppGik/motionplanners/elements/hppGikWalkElement.h"
#include "hppGik/motionplanners/hppGikLocomotionPlan.h"

#define M3_IJ MAL_S3x3_MATRIX_ACCESS_I_J
#define M4_IJ MAL_S4x4_MATRIX_ACCESS_I_J

using namespace boost::numeric::ublas;

void ChppLosMotionGenerator::createHumanoidRobot()
{
    CjrlRobotDynamicsObjectConstructor<
    dynamicsJRLJapan::DynamicMultiBody,
    Chrp2OptHumanoidDynamicRobot,
    dynamicsJRLJapan::JointFreeflyer,
    dynamicsJRLJapan::JointRotation,
    dynamicsJRLJapan::JointTranslation,
    dynamicsJRLJapan::Body> jrlRobotFactory;

    attRobot = jrlRobotFactory.createhumanoidDynamicRobot();

    aHDMB = ( dynamicsJRLJapan::HumanoidDynamicMultiBody* ) attRobot;

    std::string path = "./";
    std::string name = "HRP2.wrl";
    aHDMB->parserVRML ( path,name,"./HRP2LinkJointRank.xml" );
    std::string aName="./HRP2Specificities.xml";
    aHDMB->SetHumanoidSpecificitiesFile ( aName );
    aHDMB->SetTimeStep ( attSamplingPeriod );
    aHDMB->setComputeCoM ( true );


    aHDMB->setComputeVelocity ( false );
    aHDMB->setComputeMomentum ( false );
    aHDMB->setComputeAcceleration ( false );
    aHDMB->setComputeZMP ( false );
    aHDMB->setComputeSkewCoM ( false );
    aHDMB->setComputeAccelerationCoM ( false );
    aHDMB->setComputeBackwardDynamics ( false );



    unsigned int nDof = attRobot->numberDof();
    vectorN halfsittingConf ( nDof );

    //Half sitting
    double dInitPos[40] =
    {
        0.0, 0.0, -26.0, 50.0, -24.0, 0.0, // right leg

        0.0, 0.0, -26.0, 50.0, -24.0, 0.0, // left leg

        0.0, 0.0, // chest

        0.0, 0.0, // head

        15.0, -10.0, 0.0, -30.0, 0.0, 0.0, // right arm

        10.0, // right hand clench

        15.0,  10.0, 0.0, -30.0, 0.0, 0.0, // left arm

        10.0, // left hand clench

        -10.0, 10.0, -10.0, 10.0, -10.0, // right hand parallel mechanism
        -10.0, 10.0, -10.0, 10.0, -10.0  // left hand parallel mechanism
    };

    //waist x y z
    halfsittingConf ( 0 ) = 0.0;
    halfsittingConf ( 1 ) = 0.0;
    halfsittingConf ( 2 ) = 0.6487;
    //waist roll pitch yaw
    halfsittingConf ( 3 ) = 0.0;
    halfsittingConf ( 4 ) = 0.0;
    halfsittingConf ( 5 ) = 0.0;

    //joints
    for ( unsigned int i=6;i<nDof;i++ )
        halfsittingConf ( i ) = dInitPos[i-6]*M_PI/180;

    zero_vector<double> zeros ( attRobot->numberDof() );
    attRobot->currentConfiguration ( halfsittingConf );
    attRobot->currentVelocity ( zeros );
    attRobot->currentAcceleration ( zeros );
    attRobot->computeForwardKinematics();

    //set gaze origin and direction
    vector3d gazeDir,gazeOrigin;
    gazeDir[0] = 1;
    gazeDir[1] = 0;
    gazeDir[2] = 0;

    gazeOrigin[0] = 0;
    gazeOrigin[1] = 0;
    gazeOrigin[2] = 0.118;

    aHDMB->gaze ( ( const vector3d& ) gazeDir, ( const vector3d& ) gazeOrigin );

    ChppGik2DShape leftFootShape, rightFootShape;

    ChppGik2DVertex vert;

    vert.x = 0.135;
    vert.y = 0.079;
    leftFootShape.vertices.push_back ( vert );
    vert.x = -0.105;
    leftFootShape.vertices.push_back ( vert );
    vert.y = -0.059;
    leftFootShape.vertices.push_back ( vert );
    vert.x = 0.135;
    leftFootShape.vertices.push_back ( vert );

    vert.x = 0.135;
    vert.y = 0.059;
    rightFootShape.vertices.push_back ( vert );
    vert.x = -0.105;
    rightFootShape.vertices.push_back ( vert );
    vert.y = -0.079;
    rightFootShape.vertices.push_back ( vert );
    vert.x = 0.135;
    rightFootShape.vertices.push_back ( vert );

    attStandingRobot = new ChppGikStandingRobot ( *attRobot, leftFootShape, rightFootShape );

    attStandingRobot->staticState ( halfsittingConf );
}

ChppLosMotionGenerator::ChppLosMotionGenerator() : attSamplingPeriod ( 5e-3 )
{
    createHumanoidRobot();

    ChppLosFootprint iLeft,iRight;
    iRight.x = 0.0;
    iRight.y = -0.095;
    iRight.th = 0.0;

    iLeft.x = 0.0;
    iLeft.y = 0.095;
    iLeft.th = 0.0;

    ChppLosFootprintShape lshape;
    lshape.xDownLeft = -0.105;
    lshape.xDownRight = -0.105;
    lshape.xUpRight = 0.135;
    lshape.xUpLeft = 0.135;
    lshape.yDownLeft = 0.079;
    lshape.yUpLeft = 0.079;
    lshape.yDownRight = -0.059;
    lshape.yUpRight = -0.059;

    ChppLosFootprintArea larea ;
    larea.maxX = 0.22;
    larea.minX = -0.22;
    larea.maxY = 0.25;
    larea.minY = 0.07;
    larea.maxTheta = M_PI/4;
    larea.minTheta = -0.1;

    ChppLosFootprintShape rshape;
    rshape.xDownLeft = -0.105;
    rshape.xDownRight = -0.105;
    rshape.xUpRight = 0.135;
    rshape.xUpLeft = 0.135;
    rshape.yDownLeft = 0.059;
    rshape.yUpLeft = 0.059;
    rshape.yDownRight = -0.079;
    rshape.yUpRight = -0.079;

    ChppLosFootprintArea rarea ;
    rarea.maxX = larea.maxX;
    rarea.minX = larea.minX;
    rarea.maxY = -larea.minY;
    rarea.minY = -larea.maxY;
    rarea.maxTheta = -larea.minTheta;
    rarea.minTheta = -larea.maxTheta;

    attNSteps = 0;

    attCentipede = ChppLosCentipedeFactory::createSpringman ( iLeft,iRight, *attStandingRobot, attNSteps, false, larea,  rarea, lshape, rshape, attCentipedeParts, attMask );

    ChppLosRobotLink& Rlink = * ( * ( attCentipede->robotLinks().begin() ) );
    CjrlJoint* cJoint;
    if ( attCentipede->linkAtRank(0)->rootIsRight() )
        cJoint = attRobot->leftFoot();
    else
        cJoint = attRobot->rightFoot();

    footT = new ChppLosTransformationTask ( *attCentipede,Rlink, *cJoint, vector3d ( 0,0,0 ),cJoint->currentTransformation() );
    attCentipedeConstraints.push_back ( footT );

    comT = new ChppLosComTask ( *attCentipede,Rlink,attRobot->positionCenterOfMass() [0],attRobot->positionCenterOfMass() [1] );
    attCentipedeConstraints.push_back ( comT );

    attMotion = new ChppRobotMotion ( attRobot,0.0,attSamplingPeriod );

    attSolver = new ChppLosSolverBasic ( *attCentipede );
    attSolver->setMask(attMask);
}

bool ChppLosMotionGenerator::setSolverMask(const vectorN& inVec)
{
    if (inVec.size()!=attCentipede->numberDof())
    {
        std::cout << "pfff number of dofs not equal to entered mask size"<< std::endl;
        return false;
    }
    attMask = inVec;
    attSolver->setMask(attMask);
    return true;
}

ChppLosMotionGenerator::~ChppLosMotionGenerator()
{
    delete attStandingRobot;
    delete attCentipede;
    for ( unsigned int i=0;i<attCentipedeParts.size();i++ )
    {
        delete * ( attCentipedeParts.begin() );
        attCentipedeParts.pop_front();
    }
    for ( unsigned int i = 0;i <attCentipedeConstraints.size();i++ )
    {
        delete * ( attCentipedeConstraints.begin() );
        attCentipedeConstraints.pop_front();
    }
    delete attMotion;
    delete attSolver;
}



void ChppLosMotionGenerator::resetInitialSP()
{
    const matrix4d& lm = attRobot->leftFoot()->currentTransformation();
    const matrix4d& rm = attRobot->rightFoot()->currentTransformation();

    ChppLosFootprint iLeft,iRight;

    matrixNxP um ( 4,4 );
    vectorN euler ( 3 );
    um.clear();
    ChppGikTools::Matrix4toUblas ( lm,um );
    ChppGikTools::RottoEulerZYX ( um,euler );
    double angle = euler ( 2 );

    iLeft.x = M4_IJ ( lm,0,3 );
    iLeft.y = M4_IJ ( lm,1,3 );
    iLeft.th = angle;

    um.clear();
    ChppGikTools::Matrix4toUblas ( rm,um );
    ChppGikTools::RottoEulerZYX ( um,euler );
    angle = euler ( 2 );

    iRight.x = M4_IJ ( rm,0,3 );
    iRight.y = M4_IJ ( rm,1,3 );
    iRight.th = angle;

    attCentipede->initialLeftFootprint ( iLeft );
    attCentipede->initialRightFootprint ( iRight );
}


void ChppLosMotionGenerator::constraints ( std::vector<ChppLosTask*>& inConstraints )
{
    attConstraints = inConstraints;
}
void ChppLosMotionGenerator::tasks ( std::vector<ChppLosTask*>& inTasks )
{
    attTasks = inTasks;
}


ChppRobotMotion& ChppLosMotionGenerator::motion()
{
    return *attMotion;
}

void ChppLosMotionGenerator::generate ( ChppLosCentipede& inCentipede, const vectorN initialConfig )
{

    attStandingRobot->staticState ( initialConfig );

    //plan locomotion trajectories over time
    std::vector<ChppGikStepTarget*> walkpath;
    ChppGikFootprint footprint ( 0,0,0 );

    unsigned int curInd = 0;
    const ChppLosFootprint* fp;
    for ( unsigned int i = 0; i<inCentipede.numberLinks();i++ )
    {
        curInd = 3*i;

        if ( inCentipede.linkAtRank ( i )->rootIsRight() )
            fp = &inCentipede.linkAtRank ( i )->leftprint();
        else
            fp = &inCentipede.linkAtRank ( i )->rightprint();


        footprint.x ( fp->x );
        footprint.y ( fp->y ) ;
        footprint.th ( fp->th );
        walkpath.push_back ( new ChppGikStepTarget ( footprint, !inCentipede.linkAtRank ( i )->rootIsRight() ) );
    }

    ChppLosRobotLink* rlink =  * ( inCentipede.robotLinks().begin() );
    const vector3d& finalCom = rlink->standingRobot().robot()->positionCenterOfMass();

    double drlx = finalCom[0] - rlink->rightprint().x;
    double drly = finalCom[1] - rlink->rightprint().y;

    double cs = cos ( rlink->rightprint().th );
    double ss = sin ( rlink->rightprint().th );
    double relzmpx = cs*drlx+ss*drly;
    double relzmpy = -ss*drlx+cs*drly;

    ChppGikWalkElement* walkElement = new ChppGikWalkElement ( relzmpx, relzmpy, 0.5, attRobot, attSamplingPeriod,1.6,walkpath );
    ChppGikMotionPlan mPlan ( attRobot );
    ChppGikLocomotionPlan locoPlan ( &mPlan, attStandingRobot, attSamplingPeriod );
    locoPlan.extraEndTime ( 1.0 );
    locoPlan.addElement(walkElement);
    bool ok = locoPlan.solve();
    if ( !ok )
    {
        std::cout << "problem planning locomotion" << std::endl;
        for ( unsigned int i=0;i<walkpath.size();i++ )
            delete walkpath[i];
        return;
    }


    //determine start end end times
    double loco_end_time = locoPlan.endTime() + locoPlan.extraEndTime() + attSamplingPeriod/2;
    double loco_current_time = locoPlan.startTime() + attSamplingPeriod;
    double current_time = loco_current_time;
    double end_time = loco_end_time;
    for ( unsigned int i = 0;i<attTasks.size();i++ )
        if ( attTasks[i]->endTime() +attSamplingPeriod/2 > end_time )
            end_time = attTasks[i]->endTime() +attSamplingPeriod/2;
    /*
    for ( unsigned int i = 0;i<attConstraints.size();i++ )
        if ( attConstraints[i]->endTime() +attSamplingPeriod/2 > end_time )
            end_time = attConstraints[i]->endTime() +attSamplingPeriod/2;
    */


    //clear motion + misc
    attMotion->clear();
    std::vector<ChppLosTask*> vt;
    vectorN veloSolution ( attCentipede->numberDof() );

    while ( current_time < end_time )
    {
        //set locomotion tasks according to time
        if ( current_time < loco_end_time )
        {
            ChppGikLocomotionData* data = locoPlan.dataAtTime ( current_time );
            const vectorN& comv = data->comConstraint->worldTarget();
            const matrix4d& footm = data->footConstraint->targetTransformation();
            bool rootRight = ( data->supportJoint == attRobot->rightFoot() ) ?true:false;
            rlink =  * ( attCentipede->robotLinks().begin() );
            if ( rootRight != attCentipede->linkAtRank ( 0 )->rootIsRight() )
            {
                resetInitialSP();
                rlink->rootIsRight ( rootRight );
            }
            CjrlJoint* nsFoot = rootRight?attRobot->leftFoot() :attRobot->rightFoot();
            comT->targetXY ( comv[0], comv[1] );
            footT->joint ( nsFoot );
            footT->targetTransformation ( footm );
        }

        //gather activated tasks and constraints
        vt.clear();
        vt.push_back ( footT );
        vt.push_back ( comT );
        for ( unsigned int i=0;i<attTasks.size();i++ )
            if (( current_time > attTasks[i]->startTime() +attSamplingPeriod/2 ) && ( current_time <= attTasks[i]->endTime() +attSamplingPeriod/2 ))
                vt.push_back ( attTasks[i] );

        attSolver->tasks ( vt );
        vt.clear();
        for ( unsigned int i=0;i<attConstraints.size();i++ )
            if (( current_time > attConstraints[i]->startTime() +attSamplingPeriod/2 ) && ( current_time <= attConstraints[i]->endTime() +attSamplingPeriod/2 ))
                vt.push_back ( attConstraints[i] );
        attSolver->constraints ( vt );

        //solve constraint& task stack
        veloSolution.clear();
        attSolver->solve ( veloSolution );
        attCentipede->applyConfiguration ( veloSolution + rlink->currentConfiguration(), false );
        attStandingRobot->updateRobot ( attRobot->rootJoint()->currentTransformation(), rlink->currentConfiguration(), attSamplingPeriod );

        attMotion->appendSample ( attRobot->currentConfiguration(),vector3d(),vector3d(),vector3d(),vector3d() );

        current_time += attSamplingPeriod;
    }


    for ( unsigned int i=0;i<walkpath.size();i++ )
        delete walkpath[i];
    return;
}

