/*
  Copyright 2009 CNRS-LAAS

  Author: Oussama Kanoun
*/

#ifndef HPP_LOS_MOTION_GENERATOR_H
#define HPP_LOS_MOTION_GENERATOR_H



#include <iostream>
#include "MatrixAbstractLayer/MatrixAbstractLayer.h"
#include <boost/numeric/bindings/traits/ublas_matrix.hpp>
#include <boost/numeric/bindings/lapack/gesvd.hpp>
#include "robotDynamics/jrlRobotDynamicsObjectConstructor.h"
#include "hrp2Dynamics/hrp2OptHumanoidDynamicRobot.h"

#include "hppGik/hppGikTools.h"
#include "hppGik/robot/hppRobotMotion.h"

#include "hpplocalstepper/hppLosCentipede.hh"
#include "hpplocalstepper/tasks/hppLosTask.hh"

#include "hpplocalstepper/tasks/hppLosComBoundingTask.hh"
#include "hpplocalstepper/tasks/hppLosComTask.hh"
#include "hpplocalstepper/tasks/hppLosFeetProjectionTask.hh"
#include "hpplocalstepper/tasks/hppLosFeetAntiCollisionTask.hh"
#include "hpplocalstepper/tasks/hppLosFeetAreaTask.hh"
#include "hpplocalstepper/tasks/hppLosDiscAvoider.hh"
#include "hpplocalstepper/tasks/hppLosPositionTask.hh"
#include "hpplocalstepper/tasks/hppLosParallelTask.hh"
#include "hpplocalstepper/tasks/hppLosLinkConfigurationTask.hh"
#include "hpplocalstepper/tasks/hppLosGazeTask.hh"
#include "hpplocalstepper/hppLosSolver.hh"
#include "hpplocalstepper/hppLosCentipedeFactory.hh"
#include "hpplocalstepper/tasks/HRP2BasicCollisionAvoider.hh"
#include "hpplocalstepper/hppLosSolverBasic.hh"


class ChppLosMotionGenerator
{
    public:

        ChppLosMotionGenerator();

        ~ChppLosMotionGenerator();

        void constraints(std::vector<ChppLosTask*>& constraints);
        void tasks(std::vector<ChppLosTask*>& tasks);
        
        void generate(ChppLosCentipede& inCentipede, const vectorN initialConfig);
        ChppRobotMotion& motion();
        ChppLosCentipede& centipede(){return *attCentipede;}
        ChppLosRobotLink& robotLink(){return **attCentipede->robotLinks().begin();}
        
        bool setSolverMask(const vectorN& inVec);
        
        ChppGikStandingRobot& standingRobot(){return *attStandingRobot;}
        CjrlHumanoidDynamicRobot& robot(){return *attRobot;}
    
    private:
        void resetInitialSP();
        void createHumanoidRobot();

        ChppGikStandingRobot* attStandingRobot;

        CjrlHumanoidDynamicRobot* attRobot;
        
        ChppLosSolverBasic* attSolver;

        double attSamplingPeriod;
    
        ChppLosCentipede* attCentipede;
    
        std::list<ChppLosLink*> attCentipedeParts;
    
        std::list<ChppLosTask*> attCentipedeConstraints;
        std::vector<ChppLosTask*> attTasks,attConstraints;
    
        vectorN attMask;
    
        unsigned int attNSteps;
    
        dynamicsJRLJapan::HumanoidDynamicMultiBody *aHDMB;
    
        
        ChppLosTransformationTask* footT;
        ChppLosComTask* comT;
        ChppRobotMotion* attMotion;
};

#endif

